module Core where



import Data.List

import Names
import Pretty



-- So we can write things like (TypeKind `FunctionKind` TypeKind `FunctionKind` TypeKind) and get
-- the expected (* -> (* -> *)) instead of ((* -> *) -> *)
infixr 9 `FunctionKind`

data Strictness = Strict | Lazy deriving Eq


{-
Core.Expr describes an expression after desugaring and after type inference. All of the types on a
Core.Expr are either explicitly marked or trivial to infer. The function exprType gives the type of
a Core.Expr.
-}

data Expr
	= VarExpr (Name, FlavorType)   -- (var name, type and flavor)
	| ConExpr (CName, FlavorType)   -- (constructor name, type and flavor)
	| NumericLiteralExpr Double Type   -- value, flavor
	| StringLiteralExpr String Type   -- value, flavor
	| LambdaExpr Name FieldType Expr   -- arg, argtype, body
	| ApplyExpr Strictness Expr Expr   -- function, arg
	| CaseExpr Expr [(Pattern, Expr)]   -- value being examined, patterns and branches
	| DoExpr [Stmt] Expr   -- statements, 'then'-value

data Pattern
	= VarPattern (Name, FieldType)
	| ConApplyPattern (CName, Type) [(Strictness, Pattern)]   -- The Type is the return type of the instantiated constructor
	| WildcardPattern

data Stmt
	= BindStmt Name FieldType Expr
	| ActStmt Expr

data Type
	= VarType Name Kind
	| ConType CName Kind
	| ApplyType Type Type
	| FunctionType FieldType FlavorType
	deriving (Eq)

data Kind
	= TypeKind
	| FunctionKind Kind Kind
	deriving (Eq, Show)
monadKind = TypeKind `FunctionKind` TypeKind
transformerKind = monadKind `FunctionKind` TypeKind `FunctionKind` TypeKind

data FieldType
    = StrictField Type
    | LazyField Type Type   -- First is flavor, second is type
    deriving (Eq)

data PolyType
	= PolyType [(Name, Kind)] Type

data FlavorType
	= FlavorType Type Type   -- First is flavor, second is type
	deriving (Eq)

data PolyFlavorType
	= PolyFlavorType [(Name, Kind)] FlavorType



exprType :: Expr -> Type
exprType (VarExpr (_, FlavorType _ t)) = t
exprType (ConExpr (_, FlavorType _ t)) = t
exprType (NumericLiteralExpr _ _) = ConType (Name "Number") TypeKind
exprType (StringLiteralExpr _ _) = ConType (Name "String") TypeKind
exprType (LambdaExpr _ a b) = FunctionType a (FlavorType (exprFlavor b) (exprFlavor b))
exprType (ApplyExpr _ f _) = case exprType f of
	FunctionType _ (FlavorType _ r) -> r
	VarType _ _ -> error "exprType: incompletely typed expression"
	_ -> error "exprType: badly typed expression"
exprType (CaseExpr _ ((_,b):_)) = exprType b
exprType (DoExpr _ t) = exprType t

exprFlavor :: Expr -> Type
exprFlavor (VarExpr (_, FlavorType f _)) = f
exprFlavor (ConExpr (_, FlavorType f _)) = f
exprFlavor (NumericLiteralExpr _ f) = f
exprFlavor (StringLiteralExpr _ f) = f
exprFlavor (LambdaExpr _ _ b) = exprFlavor b
exprFlavor (ApplyExpr _ f _) = exprFlavor f
exprFlavor (CaseExpr s _) = exprFlavor s
exprFlavor (DoExpr _ t) = exprFlavor t

typeKind :: Type -> Kind
typeKind (VarType _ k) = k
typeKind (ConType _ k) = k
typeKind (ApplyType f _) = case typeKind f of
	FunctionKind _ r -> r
	_ -> error "typeKind: badly kinded type"
typeKind (FunctionType _ _) = TypeKind



instance Pretty Strictness where
    pretty Strict = "strict"
    pretty Lazy = "lazy"

instance Pretty Expr where
	pretty (VarExpr (name,ty)) = pretty name
	pretty (ConExpr (name,ty)) = pretty name
	pretty (NumericLiteralExpr n _) = show n
	pretty (StringLiteralExpr s _) = show s
	pretty (LambdaExpr name (StrictField _) body) = "(" ++ pretty name ++ " -> " ++ pretty body ++ ")"
	pretty (LambdaExpr name (LazyField _ _) body) = "([" ++ pretty name ++ "] -> " ++ pretty body ++ ")"
	pretty (ApplyExpr Strict fun arg) = "(" ++ pretty fun ++ " " ++ pretty arg ++ ")"
	pretty (ApplyExpr Lazy fun arg) = "(" ++ pretty fun ++ " [" ++ pretty arg ++ "])"
	pretty (CaseExpr subj branches) = "(case " ++ pretty subj ++ " of " ++ concat ["(" ++ pretty pat ++ " -> " ++ pretty exp ++ ")" | (pat,exp) <- branches] ++ ")"
	pretty (DoExpr stmts result) = "(do " ++ concat [pretty x ++ "; " | x <- stmts] ++ "then " ++ pretty result ++ ")"

instance Pretty Pattern where
	pretty (VarPattern (name, ty)) = pretty name ++ "::" ++ pretty ty
	pretty (ConApplyPattern (con, ty) pats) = "(" ++ pretty con ++
		concat [case strictness of
			Strict -> " " ++ pretty pat
			Lazy -> " [" ++ pretty pat ++ "]"
			| (strictness, pat) <- pats] ++ ")"
	pretty (WildcardPattern) = "_"

instance Pretty Stmt where
	pretty (BindStmt name (StrictField _) value) = pretty name ++ " = " ++ pretty value
	pretty (BindStmt name (LazyField _ _) value) = pretty name ++ " =[] " ++ pretty value
	pretty (ActStmt value) = pretty value

instance Pretty Type where
	pretty (VarType name _) = pretty name
	pretty (ConType name _) = pretty name
	pretty (ApplyType (ApplyType (ConType (Name "ComposeType") _) t) m) =
		"(" ++ pretty t ++ " / " ++ pretty m ++ ")"
	pretty (ApplyType fun arg) = "(" ++ pretty fun ++ " " ++ pretty arg ++ ")"
	pretty (FunctionType arg ret) = "(" ++ pretty arg ++ " -> " ++ pretty ret ++ ")"

instance Pretty Kind where
	pretty (TypeKind) = "*"
	pretty (FunctionKind arg ret) = "(" ++ pretty arg ++ " -> " ++ pretty ret ++ ")"

instance Pretty FieldType where
    pretty (StrictField t) = pretty t
    pretty (LazyField m t) = "[ {" ++ pretty m ++ "} " ++ pretty t ++ "]"

instance Pretty FlavorType where
	pretty (FlavorType m t) = "{" ++ pretty m ++ "} " ++ pretty t

instance Pretty PolyFlavorType where
	pretty (PolyFlavorType vars ty) = "(" ++ intercalate ", " [pretty name ++ "::" ++ pretty kind | (name,kind) <- vars] ++ ") => " ++ pretty ty
