module DesugarExpr where



import Control.Monad
import Data.Map (Map)
import qualified Data.Map as Map

import Core
import DesugarType
import ErrorHandling
import Names
import Pretty
import Scope
import Substitute
import qualified Syntax as S
import TypeInference


{-
desugarExpr converts a Syntax.Expr into a Core.Expr. The type annotations on the Core.Expr are in
terms of type variables; desugarExpr also produces a list of free type variables and a set of
constraints. The compiler solves the constraints for the free type variables, and then replaces the
type variables in the Core.Expr with the actual types.
-}



desugarExpr :: Scope
            -> S.Expr
            -- Return value is the type of the desugared expression and the desugared expression
            -- itself
            -> ConstraintMonad (FlavorType, Expr)

desugarExpr scope (S.VarExpr span name) = do

	-- Look the name up in local and global scopes to see what its type is
	polyType <- case Map.lookup name (scopeValues scope) of
		Just v -> return (pvaluePolyType v)
		Nothing -> errorLocC ("At "++pretty span++":") $ fail ("Name '"++pretty name++"' not in scope.")
	
	-- Instantiate any polymorphic variables in the type
	instType <- instantiateFlavorType ("Type of '" ++ pretty name ++"' at "++pretty span) polyType
	
	return (instType, VarExpr (name, instType))

desugarExpr scope (S.ConExpr span name) = do
	
	-- Look up the constructor to see what its type is
	con <- case Map.lookup name (scopeCons scope) of
		Just con -> return con
		Nothing -> errorLocC ("At "++pretty span++":") $ fail ("Constructor '"++pretty name++"' not in scope.")
	
	-- Instantiate any polymorphic variables in the constructor type
	instType <- instantiateFlavorType
		("Type of constructor '" ++ pretty name ++"' at " ++ pretty span)
		(conType con)
		
	return (instType, ConExpr (name, instType))

desugarExpr scope e@(S.NumericLiteralExpr span n) = do
	
	-- Numeric literals can take on any flavor
	flavor <- newFreeVar monadKind ("Flavor of numeric literal '" ++ pretty e ++ "' at " ++ pretty span)
	
	return (FlavorType flavor (ConType (Name "Number") TypeKind), NumericLiteralExpr n flavor)

desugarExpr scope e@(S.StringLiteralExpr span n) = do
	
	-- String literals can also take on any flavor
	flavor <- newFreeVar monadKind ("Flavor of string literal " ++ pretty e ++ " at " ++ pretty span)
	
	return (FlavorType flavor (ConType (Name "String") TypeKind), StringLiteralExpr n flavor)

desugarExpr scope fun@(S.LambdaExpr span args body) = do

	-- Make some variables to represent the types of the arguments
	argTypes <- sequence [do
		ty <- newFreeVar TypeKind ("Type of argument '" ++ pretty arg ++ "' in function at " ++
			pretty span)
		case strictness of
			Strict -> return (StrictField ty)
			Lazy -> do
				fl <- newFreeVar monadKind ("Flavor of argument '" ++ pretty arg ++ "' in function \
					\at " ++ pretty span)
				return (LazyField fl ty)
		| (strictness, arg) <- args]
	
	-- Insert arguments into scope
	let scope' = scope {
		scopeValues =
			Map.fromList (zip (map snd args) (map makePolyValue argTypes)) `Map.union` (scopeValues scope)
		}
		
	(FlavorType flavor bodyType, body') <-
		errorLocC ("In function at "++pretty span++":") $
			desugarExpr scope' body
	
	let functionType = foldr (\arg ret -> FunctionType arg (FlavorType flavor ret)) bodyType argTypes
	let function = foldr (\((_, name), ty) body -> LambdaExpr name ty body) body' (zip args argTypes)
	return (FlavorType flavor functionType, function)

desugarExpr scope e@(S.ApplyExpr span Strict fun arg) = do
	
	-- In strict function application, the flavor of the argument must match the flavor of the
	-- function and of the expression as a whole.
	
	(FlavorType flavor funType, fun') <-
		errorLocC ("In strict function applied at "++pretty span++":") $
			desugarExpr scope fun
	
	(FlavorType argFlavor argType, arg') <-
		errorLocC ("In argument of strict function application at "++pretty span++":") $
			desugarExpr scope arg
	constrainEqual argFlavor flavor ("flavor of '" ++ pretty arg ++ "'")
	
	retType <- newFreeVar TypeKind ("Type of expression '" ++ pretty e ++ "' at " ++ pretty span)
	constrainEqual (FunctionType (StrictField argType) (FlavorType flavor retType)) funType
		("type of expression '" ++ pretty fun ++ "' at " ++ pretty (S.spanOfExpr fun))
	
	return (FlavorType flavor retType, ApplyExpr Strict fun' arg')

desugarExpr scope e@(S.ApplyExpr span Lazy fun arg) = do
	
	-- In lazy function application, the flavor of the argument must match the flavor of the
	-- function type's argument.
	
	(FlavorType flavor funType, fun') <-
		errorLocC ("In lazy function applied at "++pretty span++":") $
			desugarExpr scope fun
	
	(FlavorType argFlavor argType, arg') <-
		errorLocC ("In argument of lazy function application at "++pretty span++":") $
			desugarExpr scope arg
	
	retType <- newFreeVar TypeKind ("Type of expression '" ++ pretty e ++ "' at " ++ pretty span)
	constrainEqual (FunctionType (LazyField argFlavor argType) (FlavorType flavor retType)) funType
		("type of expression '" ++ pretty fun ++ "' at " ++ pretty (S.spanOfExpr fun))
	
	return (FlavorType flavor retType, ApplyExpr Lazy fun' arg')

desugarExpr scope e@(S.CaseExpr span subject branches) = do
	
	-- "subject" is a term I made up to represent the expression being examined in the 'case'
	-- statement. In 'case a of b -> c', the expression 'a' is the subject.
	
	(FlavorType flavor subjType, subject') <-
		errorLocC ("In subject of case expression at "++pretty span++":") $
			desugarExpr scope subject
	
	-- Used for making sure that all branches have the same type
	resultType <- newFreeVar TypeKind ("Type of expression '" ++ pretty e ++ "' at " ++ pretty span)
	
	branches' <- sequence [do
		
		(pattern', varTypes) <-
			errorLocC ("In pattern #"++show n++" of case expression at "++pretty span++":") $
				desugarPattern scope (StrictField subjType) flavor pattern
		
		-- Insert any new variables defined by the pattern into scope
		let scope' = scope { scopeValues = Map.map makePolyValue varTypes `Map.union` (scopeValues scope) }
		
		(FlavorType expFlavor expType, exp') <-
			errorLocC ("In branch #"++show n++" of case expression at "++pretty span++":") $
				desugarExpr scope' exp
		
		-- Each branch of the 'case' must have the same type and flavor
		constrainEqual expFlavor flavor ("flavor of '" ++ pretty exp ++ "'")
		constrainEqual expType resultType ("type of '" ++ pretty exp ++ "'")
		
		return (pattern', exp')
		
		| ((pattern, exp), n) <- zip branches [1..]]
	
	return (FlavorType flavor resultType, CaseExpr subject' branches')

desugarExpr scope e@(S.DoExpr span stmts result) = errorLocC ("In 'do' expression at "++pretty span++":") $ do
	
	flavor <- newFreeVar monadKind ("flavor of 'do' expression at "++pretty span)
	
	let
		handleSequence :: Scope
		               -> [S.Stmt]
		               -> ConstraintMonad ([Stmt], Type, Expr)
		
		handleSequence scope ((S.BindStmt span strictness name value):rest) = do
		
			(FlavorType valueFlavor valueType, value') <- desugarExpr scope value
			
			ty <- case strictness of
				Strict -> do
					-- In a strict binding (which behaves like do-notation in Haskell) the flavor
					-- of the binding must be the same as the flavor of the entire do-block.
					constrainEqual valueFlavor flavor ("flavor of '" ++ pretty value ++ "'")
					return (StrictField valueType)
				Lazy -> do
					-- In a lazy binding (which behaves like let-notation in Haskell) the flavor
					-- of the binding can be anything.
					return (LazyField valueFlavor valueType)
			
			let scope' = scope {
				scopeValues = Map.insert name (makePolyValue ty) (scopeValues scope)
				}
			let stmt' = BindStmt name ty value'
			
			(stmts', resultType, result') <- handleSequence scope' rest
			return (stmt':stmts', resultType, result')
		
		handleSequence scope ((S.ActStmt span value):rest) = do
		
			(FlavorType valueFlavor _, value') <- desugarExpr scope value
			constrainEqual valueFlavor flavor ("flavor of '" ++ pretty value ++ "'")
			
			let stmt' = ActStmt value'
			
			(stmts', resultType, result') <- handleSequence scope rest
			return (stmt':stmts', resultType, result')
		
		handleSequence scope [] = do
		
			(FlavorType resultFlavor resultType, result') <- desugarExpr scope result
			constrainEqual resultFlavor flavor ("flavor of '" ++ pretty result ++ "'")
			
			return ([], resultType, result')
	
	(stmts', resultType, result') <- handleSequence scope stmts
	
	return (FlavorType flavor resultType, DoExpr stmts' result')



desugarPattern :: Scope
               -- The type that the pattern is expected to have
               -> FieldType
               -- The flavor that the pattern is being evaluated in (lazy fields can only be
               -- matched against if their flavor is the same as the flavor being evaluated in)
               -> Type
               -- The pattern to desugar
               -> S.Pattern
               -- Return value is the desugared form of the pattern and a map with the types of all
               -- the variables in the pattern
               -> ConstraintMonad (Pattern, Map Name FieldType)

desugarPattern scope ty flavor (S.VarPattern span name) = do
	
	return (VarPattern (name, ty), Map.singleton name ty)

desugarPattern scope ty flavor (S.ConApplyPattern span name fields) =

	case Map.lookup name (scopeCons scope) of
	
		Just con -> do
			
			-- Instantiate any polymorphic variables in the type of the constructor
			subs <- instantiateFreeVars ("Type of constructor '"++pretty name++"' at "++pretty span)
				(conFreeVars con)
			
			-- Make sure we are giving the constructor the right number of fields
			unless (length (conFieldTypes con) == length fields) $
				fail ("Constructor '"++pretty name++"' has "++show (length (conFieldTypes con))++
					" field(s), but you are providing "++show (length fields)++" pattern(s).")
			
			-- Desugar each field in the constructor
			(fields', varTypes) <- liftM unzip $ sequence [
				errorLocC ("In field #"++show place++" of constructor "++pretty name++":") $ do
				
					-- Make sure that the strictness in the pattern is the same as the strictness
					-- in the constructor's definition
					case uninstFieldType of
						StrictField _ -> unless (strictness == Strict) $
							fail ("Field #"++show place++" in constructor "++pretty name++" is \
								\strict, but you are marking it as lazy in this pattern.")
						LazyField _ _ -> unless (strictness == Lazy) $
							fail ("Field #"++show place++" in constructor "++pretty name++" is \
								\lazy, but you are marking it as strict in this pattern. Put '[]' \
								\brackets around the field in the pattern to make it as lazy.")
					
					-- Recursively desugar the field's pattern
					let instFieldType = substituteTypeInFieldType subs uninstFieldType
					(field', varTypes) <- desugarPattern scope instFieldType flavor field
					
					return ((strictness, field'), varTypes)
				
				| (place, (strictness, field), uninstFieldType) <- zip3 [1..] fields (conFieldTypes con)]
			
			-- Make sure that the constructor's return type matches
			let instResultType = substituteTypeInType subs (conReturnType con)
			case ty of
				StrictField ty -> do
					constrainEqual instResultType ty ("type of pattern")
				LazyField fl ty -> do
					constrainEqual instResultType ty ("type of pattern")
					-- If this pattern appears as a lazy field of another pattern, then monadic
					-- actions will be executed when the pattern is evaluated. This is only possible
					-- if the flavor of the pattern match as a whole is the same as the flavor of
					-- the lazy field.
					constrainEqual fl flavor ("flavor used to evaluate lazy pattern matches")
			
			return (ConApplyPattern (name, instResultType) fields', Map.unions varTypes)
		
		Nothing -> errorLocC ("At "++pretty span++":") $ fail ("Constructor '"++pretty name++"' not in scope.")

desugarPattern scope ty flavor (S.WildcardPattern span) = do
	
	return (WildcardPattern, Map.empty)



makePolyValue :: FieldType -> PolyValue
makePolyValue (StrictField ty) = PolyValue {
	pvalueFreeVars = [(Name "_f", monadKind)],
	pvalueType = FlavorType (VarType (Name "_f") monadKind) ty,
	pvalueImplExpr = Nothing,
	pvalueImplHaskell = Nothing
	}
makePolyValue (LazyField fl ty) = PolyValue {
	pvalueFreeVars = [],
	pvalueType = FlavorType fl ty,
	pvalueImplExpr = Nothing,
	pvalueImplHaskell = Nothing
	}



instantiateFreeVars :: String -> [(Name, Kind)] -> ConstraintMonad (Map Name Type)
instantiateFreeVars what kinds = do
	repls <- sequence [newFreeVar kind ("part of " ++ what) | (var, kind) <- kinds]
	return (Map.fromList (zip (map fst kinds) repls))

instantiateFlavorType :: String -> PolyFlavorType -> ConstraintMonad FlavorType
instantiateFlavorType what (PolyFlavorType polyVars (FlavorType uninstFlavor uninstType)) = do
	sub <- liftM substituteTypeInType $ instantiateFreeVars what polyVars
	return (FlavorType (sub uninstFlavor) (sub uninstType))
