module Parser where



import Control.Monad
import Control.Monad.Identity
import Text.Parsec
import Text.Parsec.String

import Syntax
import Core(Kind(..), Strictness(..))
import Names
import Pretty



parseProgram :: String -> String -> Either ParseError [Decl]
parseProgram filename source =
	let pProgram = do
		whitespace
		decls <- many $ do
			d <- pDecl
			linebreak
			return d
		eof
		return decls
	in parse pProgram filename source



type MEParser = Parser

lexeme :: MEParser x -> MEParser x
lexeme x = do
	y <- x
	whitespace
	return y

symbol :: String -> MEParser String
symbol = lexeme . string

keyword :: String -> MEParser String
keyword x = try $ lexeme $ do
	string x
	notFollowedBy $ oneOf (lowercaseChars ++ uppercaseChars ++ otherNameChars)
	return x

keysym :: String -> MEParser String
keysym x = try $ lexeme $ do
	string x
	notFollowedBy $ oneOf symbolChars
	return x

lowercaseChars = "abcdefghijklmnopqrstuvwxyz"
uppercaseChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
otherNameChars = "'0123456789"
symbolChars = "!@#$%^&*-=+|\\:<>./?~"

whitespace, linebreak, comment, blockComment :: MEParser ()
whitespace = many (comment <|> blockComment <|> wsnl) >> return ()
	where wsnl = oneOf " \t\r\n" >> return ()
linebreak = symbol ";" >> return ()
comment = do
	try $ string "--"
	many (noneOf "\r\n")
	oneOf "\r\n"
	return ()
blockComment = do
	try $ string "{-"
	many (noneOf "-" <|> try (char '-' >> noneOf "}") )
	string "-}"
	return ()

block :: MEParser a -> MEParser [a]
block elem = do
	symbol "{"
	elems <- elem `sepEndBy` linebreak
	symbol "}"
	return elems



pWord :: MEParser String
pWord = lexeme $ do
	first <- oneOf lowercaseChars
	rest <- many $ oneOf (lowercaseChars ++ uppercaseChars ++ otherNameChars)
	return (first:rest)

pName :: MEParser Name
pName = (try $ do
	word <- pWord
	when (word `elem` ["case", "of", "do", "then", "data", "transformer", "where"]) mzero
	return (Name word)
	) <?> "name"

pInfixName :: MEParser Name
pInfixName = lexeme $ try $ do
	char '`'
	first <- oneOf lowercaseChars
	rest <- many $ oneOf (lowercaseChars ++ uppercaseChars ++ otherNameChars)
	char '`'
	return (Name (first:rest))

pCName :: MEParser CName
pCName = (lexeme $ do
	first <- oneOf uppercaseChars
	rest <- many $ oneOf (lowercaseChars ++ uppercaseChars ++ otherNameChars)
	let name = first:rest
	return (Name name)
	) <?> "constructor name"

pInfixCName :: MEParser Name
pInfixCName = lexeme $ try $ do
	char '`'
	first <- oneOf uppercaseChars
	rest <- many $ oneOf (lowercaseChars ++ uppercaseChars ++ otherNameChars)
	char '`'
	return (Name (first:rest))

pSymbol :: MEParser String
pSymbol = (lexeme $ many1 $ oneOf symbolChars) <?> "symbol"

pNumber :: MEParser Double
pNumber = (lexeme $ do
	negative <- option '+' (char '-')
	let maybeNegate = case negative of
		'-' -> negate
		'+' -> id
	iPart <- many1 $ oneOf "0123456789"
	fPart <- option "" $ do
		string "."
		digits2 <- many1 $ oneOf "0123456789"
		return ("." ++ digits2)
	return (maybeNegate (read (iPart ++ fPart)))
	) <?> "number"

pStringLiteral :: MEParser String
pStringLiteral = (lexeme $ do
	char '"'
	let oneChar = noneOf "\"\\\n" <|> do
		char '\\'
		c <- anyChar
		case c of
			'r' -> return '\r'
			'n' -> return '\n'
			't' -> return '\t'
			'\"' -> return '\"'
			'\\' -> return '\\'
			_ -> fail ("Unknown string escape: " ++ show ['\\', c] ++ ".")
	content <- many oneChar
	char '"'
	return content
	)



pDecl :: MEParser Decl
pDecl = (do
		startPos <- getPosition
		keyword "data"
		name <- pCName
		args <- many $ do
			symbol "("
			argName <- pName
			keysym "::"
			argKind <- pKind
			symbol ")"
			return (argName, argKind)
		keysym "="
		(do
				ctors <- block $ do
					ctorName <- pCName
					fieldTypes <- many (pFieldType pType4)
					return (ctorName, fieldTypes)
				endPos <- getPosition
				return (DataDecl {
					declSpan = SrcSpan startPos endPos,
					declTypeName = name,
					declFreeVars = args,
					declCons = UserDefined ctors
					})
			) <|> (do
				symbol "__builtin_Haskell"
				impl <- pStringLiteral
				endPos <- getPosition
				return (DataDecl {
					declSpan = SrcSpan startPos endPos,
					declTypeName = name,
					declFreeVars = args,
					declCons = Builtin impl
					})
			)
	
	) <|> (do
		startPos <- getPosition
		keyword "transformer"
		name <- pCName
		args <- many $ do
			symbol "("
			argName <- pName
			keysym "::"
			argKind <- pKind
			symbol ")"
			return (argName, argKind)
		keyword "where"
		defns <- block $ do
			name <- pName
			keysym "="
			value <- pExpr
			return (name, value)
		endPos <- getPosition
		(r,b) <- case defns of
			[(Name "return", r), (Name "bind", b)] -> return (r,b)
			[(Name "bind", b), (Name "return", r)] -> return (r,b)
			_ -> fail ("'transformer' block should contain two definitions: 'bind' and 'return'.")
		return (TransformerDecl {
			declSpan = SrcSpan startPos endPos,
			declTransformerName = name,
			declFreeVars = args,
			declTransformerDefn = (r,b)
			})
		
	) <|> (do
		startPos <- getPosition
		name <- pName
		rest <- (do
				keysym "="
				(do
						symbol "__builtin_Haskell"
						impl <- pStringLiteral
						endPos <- getPosition
						return (DefnDecl {
							declSpan = SrcSpan startPos endPos,
							declName = name,
							declExpr = Builtin impl
							})
					) <|> (do
						value <- pExpr
						endPos <- getPosition
						return (DefnDecl {
							declSpan = SrcSpan startPos endPos,
							declName = name,
							declExpr = UserDefined value
							})
					)
			) <|> (do
				keysym "::"
				freeVars <- option [] $ do
					symbol "("
					vars <- (`sepBy` symbol ",") $ do
						var <- pName
						keysym "::"
						kind <- pKind
						return (var, kind)
					symbol ")"
					keysym "=>"
					return vars
				symbol "{"
				flavor <- pType
				symbol "}"
				ty <- pType
				endPos <- getPosition
				return (TypeSignatureDecl {
					declSpan = SrcSpan startPos endPos,
					declName = name,
					declType = PolyFlavorType freeVars (FlavorType flavor ty)
					})
			)
			
		return rest
	)



pExpr :: MEParser Expr
pExpr = (do
		-- Case statement
		startPos <- getPosition
		keyword "case"
		subj <- pExpr
		keyword "of"
		cases <- block $ do
			pat <- pPattern
			keysym "->"
			exp <- pExpr
			return (pat, exp)
		endPos <- getPosition
		return (CaseExpr (SrcSpan startPos endPos) subj cases)
		
	) <|> (do
		-- Do statement
		startPos <- getPosition
		keyword "do"
		stmts <- block pStmt
		keyword "then"
		result <- pExpr
		endPos <- getPosition
		return (DoExpr (SrcSpan startPos endPos) stmts result)
		
	) <|> (do
		-- Lambda expression
		startPos <- getPosition
		keysym "\\"
		args <- many $ (do
				symbol "["
				n <- pName
				symbol "]"
				return (Lazy, n)
			) <|> (do
				n <- pName
				return (Strict, n)
			)
		keysym "->"
		body <- pExpr
		endPos <- getPosition
		return (LambdaExpr (SrcSpan startPos endPos) args body)
		
	) <|> (do
		
		startPos <- getPosition
		left <- pExpr2
		option left $ do
			
			-- Infix application
			op <- (do
				opStartPos <- getPosition
				(exprType, name) <-
					liftM ((,) VarExpr) pInfixName <|> liftM ((,) ConExpr) pInfixCName
				opEndPos <- getPosition
				return (exprType (SrcSpan opStartPos opEndPos) name)
				)
				
			right <- pExpr
			endPos <- getPosition
			return (ApplyExpr (SrcSpan startPos endPos) Strict
				(ApplyExpr (SrcSpan startPos endPos) Strict op left)
				right)
	) <?> "expression"

pExpr2 :: MEParser Expr
pExpr2 = do
	startPos <- getPosition
	first <- pExpr3
	apps <- many $ (do
			symbol "["
			arg <- pExpr
			symbol "]"
			midPos <- getPosition
			return (Lazy, arg, midPos)
		) <|> (do
			arg <- pExpr3
			midPos <- getPosition
			return (Strict, arg, midPos)
		)
	return $ foldl (\fun (which, arg, midPos) ->
		ApplyExpr (SrcSpan startPos midPos) which fun arg)
		first apps

pExpr3 :: MEParser Expr
pExpr3 = (do
		startPos <- getPosition
		name <- pName
		endPos <- getPosition
		return (VarExpr (SrcSpan startPos endPos) name)
	) <|> (do
		startPos <- getPosition
		name <- pCName
		endPos <- getPosition
		return (ConExpr (SrcSpan startPos endPos) name)
	) <|> (do
		startPos <- getPosition
		value <- pNumber
		endPos <- getPosition
		return (NumericLiteralExpr (SrcSpan startPos endPos) value)
	) <|> (do
		startPos <- getPosition
		value <- pStringLiteral
		endPos <- getPosition
		return (StringLiteralExpr (SrcSpan startPos endPos) value)
	) <|> (do
		symbol "("
		e <- pExpr
		symbol ")"
		return e
	)

pStmt :: MEParser Stmt
pStmt = (do
		startPos <- getPosition
		(name, strictness) <- try $ do
			name <- pName
			strictness <- (keysym "=[]" >> return Lazy) <|> (keysym "=" >> return Strict)
			return (name, strictness)
		value <- pExpr
		endPos <- getPosition
		return (BindStmt (SrcSpan startPos endPos) strictness name value)
	) <|> (do
		startPos <- getPosition
		action <- pExpr
		endPos <- getPosition
		return (ActStmt (SrcSpan startPos endPos) action)
	) <?> "statement"



pPattern :: MEParser Pattern
pPattern = (do
		startPos <- getPosition
		con <- pCName
		args <- many $ (do
				symbol "["
				pat <- pPattern
				symbol "]"
				return (Lazy, pat)
			) <|> (do
				pat <- pPattern2
				return (Strict, pat)
			)
		endPos <- getPosition
		return (ConApplyPattern (SrcSpan startPos endPos) con args)
	) <|> (do
		pPattern2
	) <?> "pattern"

pPattern2 :: MEParser Pattern
pPattern2 = (do
		startPos <- getPosition
		name <- pName
		endPos <- getPosition
		return (VarPattern (SrcSpan startPos endPos) name)
	) <|> (do
		startPos <- getPosition
		con <- pCName
		endPos <- getPosition
		return (ConApplyPattern (SrcSpan startPos endPos) con [])
	) <|> (do
		startPos <- getPosition
		keysym "_"
		endPos <- getPosition
		return (WildcardPattern (SrcSpan startPos endPos))
	) <|> (do
		symbol "("
		p <- pPattern
		symbol ")"
		return p
	)
	


pType :: MEParser Type
pType = (do
		startPos <- getPosition
		first <- try $ do
			first <- pType2
			keysym "/"
			return first
		second <- pType
		endPos <- getPosition
		let span = SrcSpan startPos endPos
		let a `apply` b = ApplyType span a b
		return (ConType span (Name "ComposeType") `apply` first `apply` second)
	) <|> pType2
	<?> "type"

pType2 :: MEParser Type
pType2 = (do
		startPos <- getPosition
		argTy <- try $ do
			t <- pFieldType pType3
			keysym "->"
			return t
		flavor <- optionMaybe $ do
			symbol "{"
			f <- pType
			symbol "}"
			return f
		retTy <- pType2
		endPos <- getPosition
		return (FunctionType (SrcSpan startPos endPos) argTy flavor retTy)
	) <|> pType3

pType3 :: MEParser Type
pType3 = do
	startPos <- getPosition
	first <- pType4
	apps <- many $ do
		arg <- pType4
		midPos <- getPosition
		return (arg, midPos)
	return (foldl (\fun (arg, midPos) -> ApplyType (SrcSpan startPos midPos) fun arg) first apps)

pType4 :: MEParser Type
pType4 = (do
		startPos <- getPosition
		name <- pName
		endPos <- getPosition
		return (VarType (SrcSpan startPos endPos) name)
	) <|> (do
		startPos <- getPosition
		name <- pCName
		endPos <- getPosition
		return (ConType (SrcSpan startPos endPos) name)
	) <|> (do
		symbol "("
		t <- pType
		symbol ")"
		return t
	)



-- pFieldType is parameterized on the underlying type parser because it might need to be used with
-- either pType3 or pType4 depending on the context
pFieldType :: MEParser Type -> MEParser FieldType
pFieldType tyParser = (do
		symbol "["
		flavor <- optionMaybe $ do
			symbol "{"
			f <- pType
			symbol "}"
			return f
		ty <- pType
		symbol "]"
		return (LazyField flavor ty)
	) <|> (do
		ty <- tyParser <?> "type"
		return (StrictField ty)
	)



pKind :: MEParser Kind
pKind = (do
		argKind <- try $ do
			argKind <- pKind2
			keysym "->"
			return argKind
		retKind <- pKind
		return (FunctionKind argKind retKind)
	) <|> (do
		pKind2
	) <?> "kind"

pKind2 :: MEParser Kind
pKind2 = (do
		keysym "*"
		return TypeKind
	) <|> (do
		symbol "("
		k <- pKind
		symbol ")"
		return k
	)
