module Scope where



import Data.Map (Map)
import qualified Data.Map as Map

import Core
import Names




{-
The Scope type contains constructors, definitions, types, monad instances, etc.
-- basically, everything that can be defined in a program. The compiler starts with builtinScope
(defined in Builtins.hs) and then adds more things as it processes the input file.
-}

data Scope = Scope {

	-- For each constructor: free variables and their kinds, field types, return type
	scopeCons :: Map Name ConInfo,
	
	-- Kinds of type constructors, and whether or not they take a default flavor argument
	scopeTypes :: Map Name TypeInfo,
	
	-- The implementations of values
	scopeValues :: Map Name PolyValue,
	
	-- Monad transformer instances.
	scopeTransformers :: Map Name TransformerInfo
	}

emptyScope = Scope {
	scopeCons = Map.empty,
	scopeTypes = Map.empty,
	scopeValues = Map.empty,
	scopeTransformers = Map.empty
	}



-- A PolyValue is a polymorphic value
data PolyValue = PolyValue {

	pvalueFreeVars :: [(Name, Kind)],
	
	pvalueType :: FlavorType,
	
	-- There are several different runtimes that a PolyValue might need to run on. The pvalueImpl*
	-- fields store implementations of the PolyValue for the different runtimes. Any of them can
	-- be "Nothing".
	
	pvalueImplExpr :: Maybe Expr,
    
	pvalueImplHaskell :: Maybe String
	}

pvaluePolyType :: PolyValue -> PolyFlavorType
pvaluePolyType pv = PolyFlavorType (pvalueFreeVars pv) (pvalueType pv)



{-
Consider the following example:

	data List (a :: *) = Cons a (List a) | Nil
	
This would be represented as a TypeInfo and two ConInfo's, as:

	-- Stored under the key (Name "List")
	AlgebraicTypeInfo {
		typeInfoKind = TypeKind `FunctionKind` TypeKind,
		typeInfoCons = [Name "Cons", Name "Nil"]
	}
	
	-- Stored under the key (Name "Cons")
	ConInfo {
		conFreeVars = [(Name "a", TypeKind)],
		conFieldTypes = [VarType (Name "a"), ConType (Name "List") `ApplyType` VarType (Name "a")],
		conReturnType = ConType (Name "List") `ApplyType` VarType (Name "a")
		}
	
	-- Stored under the key (Name "Nil")
	ConInfo {
		conFreeVars = [(Name "a", TypeKind)],
		conFieldTypes = [],
		conReturnType = ConType (Name "List") `ApplyType` VarType (Name "a")
		}
-}

data TypeInfo
	= AlgebraicTypeInfo {
		typeInfoFreeVars :: [(Name, Kind)],
		typeInfoCons :: [CName]
		}
	| BuiltinTypeInfo {
		typeInfoFreeVars :: [(Name, Kind)],
		typeInfoImplHaskell :: Maybe String
		}
typeInfoKind t = foldr FunctionKind TypeKind (map snd (typeInfoFreeVars t))

data ConInfo = ConInfo {
	conFreeVars :: [(Name, Kind)],
	conFieldTypes :: [FieldType],
	conReturnType :: Type
	}
conTypeInFlavor con flavor =
	(FlavorType flavor
		(foldr
			(\fieldType retType -> FunctionType fieldType (FlavorType flavor retType))
			(conReturnType con)
			(conFieldTypes con)
		)
	)
conType con = PolyFlavorType
	((Name "_conFlavor", monadKind) : conFreeVars con)
	(conTypeInFlavor con (VarType (Name "_conFlavor") monadKind))


data TransformerInfo = TransformerInfo {
	transformerReturn :: PolyValue,
	transformerBind :: PolyValue
	}

