module Syntax where



import Text.Parsec

import Core (Kind, Strictness(..))
import Names
import Pretty



data SrcSpan = SrcSpan SourcePos SourcePos

instance Pretty SrcSpan where
	pretty (SrcSpan start end)
		| sourceName start == sourceName end = "<" ++ sourceName start ++ ">: " ++ f start ++ " - " ++ f end
		| otherwise = pretty start ++ " - " ++ pretty end ++ " (by the way, how did you manage to make a syntactic construct span multiple files?)"
		where f p = show (sourceLine p) ++ "." ++ show (sourceColumn p)

instance Pretty SourcePos where
	pretty sp = "<" ++ sourceName sp ++ ">: " ++ show (sourceLine sp) ++ "." ++ show (sourceColumn sp)



data BuiltinOr x = Builtin String | UserDefined x

data Decl
	= DefnDecl {
		declSpan :: SrcSpan,
		declName :: Name,
		declExpr :: BuiltinOr Expr
		}
	| DataDecl {
		declSpan :: SrcSpan,
		declTypeName :: CName,
		declFreeVars :: [(Name, Kind)],
		declCons :: BuiltinOr [(CName, [FieldType])]
		}
	| TypeSignatureDecl {
		declSpan :: SrcSpan,
		declName :: Name,
		declType :: PolyFlavorType
		}
	| TransformerDecl {
		declSpan :: SrcSpan,
		declTransformerName :: CName,
		declFreeVars :: [(Name, Kind)],
		declTransformerDefn :: (Expr, Expr)
		}



data Expr
	= VarExpr SrcSpan Name
	| ConExpr SrcSpan CName
	| NumericLiteralExpr SrcSpan Double
	| StringLiteralExpr SrcSpan String
	| LambdaExpr SrcSpan [(Strictness, Name)] Expr   -- "a1 a2 [a3] [a4] -> b"
	| ApplyExpr SrcSpan Strictness Expr Expr   -- "f x" or "f [x]"
	| CaseExpr SrcSpan Expr [(Pattern, Expr)]   -- "case x of ..."
	| DoExpr SrcSpan [Stmt] Expr   -- "do ... then x"

instance Pretty Expr where
	pretty (VarExpr _ name) = pretty name
	pretty (ConExpr _ name) = pretty name
	pretty (NumericLiteralExpr _ n) = show n
	pretty (StringLiteralExpr _ s) = show s
	pretty (LambdaExpr _ args body) =
		"(" ++ concat [case strictness of
			Strict -> pretty arg ++ " "
			Lazy -> "[" ++ pretty arg ++ "] "
			| (strictness, arg) <- args] ++
			"-> " ++ pretty body ++ ")"
	pretty (ApplyExpr _ Strict fun arg) = "(" ++ pretty fun ++ " " ++ pretty arg ++ ")"
	pretty (ApplyExpr _ Lazy fun arg) = "(" ++ pretty fun ++ " [" ++ pretty arg ++ "])"
	pretty (CaseExpr _ subj branches) = "(case " ++ pretty subj ++ " of " ++ concat ["(" ++ pretty pat ++ " -> " ++ pretty exp ++ ")" | (pat,exp) <- branches] ++ ")"
	pretty (DoExpr _ stmts result) = "(do " ++ concat [pretty x ++ "; " | x <- stmts] ++ "then " ++ pretty result ++ ")"

spanOfExpr :: Expr -> SrcSpan
spanOfExpr (VarExpr s _) = s
spanOfExpr (ConExpr s _) = s
spanOfExpr (NumericLiteralExpr s _) = s
spanOfExpr (StringLiteralExpr s _) = s
spanOfExpr (LambdaExpr s _ _) = s
spanOfExpr (ApplyExpr s _ _ _) = s
spanOfExpr (CaseExpr s _ _) = s
spanOfExpr (DoExpr s _ _) = s



data Type
	= VarType SrcSpan Name
	| ConType SrcSpan CName
	| ApplyType SrcSpan Type Type
	| FunctionType SrcSpan FieldType (Maybe Type) Type

instance Pretty Type where
	pretty (VarType _ name) = pretty name
	pretty (ConType _ name) = pretty name
	pretty (ApplyType _ (ApplyType _ (ConType _ (Name "ComposeType")) t) m) =
		"(" ++ pretty t ++ " / " ++ pretty m ++ ")"
	pretty (ApplyType _ fun arg) =
		"(" ++ pretty fun ++ " " ++ pretty arg ++ ")"
	pretty (FunctionType _ arg maybeFlavor ret) =
		"(" ++ pretty arg ++ " -> " ++ fl ++ pretty ret ++ ")"
		where
			fl = case maybeFlavor of
				Nothing -> ""
				Just f -> "{" ++ pretty f ++ "} "



data Pattern
	= VarPattern SrcSpan Name
	| ConApplyPattern SrcSpan CName [(Strictness, Pattern)]
	| WildcardPattern SrcSpan

instance Pretty Pattern where
	pretty (VarPattern _ name) = pretty name
	pretty (ConApplyPattern _ con pats) = "(" ++ pretty con ++
		concat [case strictness of
			Strict -> " " ++ pretty pat
			Lazy -> " [" ++ pretty pat ++ "]"
			| (strictness, pat) <- pats] ++ ")"
	pretty (WildcardPattern _) = "_"



data Stmt
	= BindStmt SrcSpan Strictness Name Expr
	| ActStmt SrcSpan Expr

instance Pretty Stmt where
	pretty (BindStmt _ Strict name value) = pretty name ++ " = " ++ pretty value
	pretty (BindStmt _ Lazy name value) = pretty name ++ " = [] " ++ pretty value
	pretty (ActStmt _ value) = pretty value



data FieldType
	= StrictField Type
	| LazyField (Maybe Type) Type

instance Pretty FieldType where
	pretty (StrictField ty) = pretty ty
	pretty (LazyField Nothing ty) = "{" ++ pretty ty ++ "}"
	pretty (LazyField (Just flavor) ty) = "[ {" ++ pretty flavor ++ "} " ++ pretty ty ++ "]"



data PolyType
	= PolyType [(Name, Kind)] Type



data FlavorType
	= FlavorType Type Type   -- First is flavor, second is type

instance Pretty FlavorType where
	pretty (FlavorType f t) = "[" ++ pretty f ++ "] " ++ pretty t



data PolyFlavorType
	= PolyFlavorType [(Name, Kind)] FlavorType

instance Pretty PolyFlavorType where
	pretty (PolyFlavorType vars ty) = concat ["(" ++ pretty name ++ "::" ++ pretty kind ++ ")" | (name,kind) <- vars] ++ " => " ++ pretty ty
